@@ -1,14 +1,16 @@
-import pyautogui, re, time, keyboard, os, shutil, click, threading # type: ignore
-from tkinter import *
+import pyautogui, re, time, os, shutil, keyboard, click, threading, pyperclip # type: ignore
 import tkinter.messagebox as tkmb
 from datetime import datetime
 from os.path import exists
 from datetime import datetime
-if os.name == 'nt': import msvcrt
+from pynput.keyboard import Key, Controller # type: ignore
+from pynput import keyboard as kbd # type: ignore
+
+kb = Controller()
 
 #-----------------------------------------------------
 #
-# Clickomat v0.3.3
+# Clickomat v1.0.2
 #
 #-----------------------------------------------------
 
@@ -29,7 +31,7 @@ class Watcher(threading.Thread):
                 # _target comes in as list
                 # _target[0]: list of images
                 # _target[1]: target section
-                if self.parent._findImage(self._target[0]):
+                if self.parent._findImages(self._target[0]):
                     return self.jumpSection()
 
         if self.name == "whitelist":
@@ -42,16 +44,16 @@ class Watcher(threading.Thread):
         if self.parent.logging: print(f"\nImage lookup found -> Go Section {self.parent.section}!\n")
         self.parent.ClickLoop(self.parent.section)
 
-    def nullTarget(self):
+    def nullTarget(self) -> None:
         self._target = []
 
-    def setTarget(self,target):
+    def setTarget(self,target) -> None:
         self._target = target
 
-    def stop(self):
+    def stop(self) -> None:
         self._stop_event.set()
 
-    def run(self):
+    def run(self) -> None:
         while not self._stop_event.is_set():
             self.check()
 # endregion
@@ -66,33 +68,52 @@ class Failsave(threading.Thread):
         self.height      = size.height
         self.width       = size.width
 
-    def check(self):
+    def check(self) -> None:
         pos = pyautogui.position()
         if (pos.x < 4 and pos.y > self.height-4 ) \
         or (pos.x > self.width-4 and pos.y > self.height-4 ) \
         or (pos.x < 4 and pos.y < 4 ) \
         or (pos.x > self.width-4 and pos.y < 4 ) :
-            self.fullPanic()
+            self.parent._panic()
 
-    def stop(self):
+    def stop(self) -> None:
         self._stop_event.set()
 
-    def fullPanic(self):
-        print("Panic Stopp!!!")
-        self.parent.error = "Panic Stopp!!!"
-        self.parent.panicked = True
-        self.parent.stopped = True
-        self.parent.breakout = True
-        self.parent._stopAllTreads()
+    def run(self) -> None:
+        while not self._stop_event.is_set():
+            self.check()
+# endregion
+# region StopLoop Thread
+class StopLoop(threading.Thread):
+    def __init__(self,parent):
+        threading.Thread.__init__(self)
+        self.parent      = parent
+        self._stop_event = threading.Event()
 
-    def run(self):
+    def check(self) -> None:
+
+        if self.parent.breakout:
+            if self.parent.logging and not self.parent.panicked: print ("Loop broken!\n\n")
+            message = "An error has occured: " + self.parent.error
+            self.parent._popupMessage(message,'error')
+            self.parent._stopAllTreads()
+            exit()
+        if self.parent.stopped:
+            if self.parent.logging: print ("Loop stopped!\n\n")
+            self.parent._stopAllTreads()
+            exit()
+
+    def stop(self) -> None:
+        self._stop_event.set()
+
+    def run(self) -> None:
         while not self._stop_event.is_set():
             self.check()
 # endregion
 
 class Clickomat:
     # region __init__
-    def __init__(self, case_path=None, input_file=None, images=None):
+    def __init__(self, case_path=None, input_file=None, targets_folder=None):
         # If ran from Python we grab the params here
         # self.commands is used if a multiline string is given instead a clicklist filename
         self.commands = None
@@ -130,12 +151,12 @@ class Clickomat:
                     exit()
 
         # images contains the name of the folder where the targets (PNGs) are stored
-        if images is None:
-            self.images = "."
+        if targets_folder is None:
+            self.targets_folder = "."
         else:
             # ... and of course we check for existence
-            self.images = f"{self.case_path}/{images}"
-            if not os.path.isdir(self.images):
+            self.targets_folder = f"{self.case_path}/{targets_folder}"
+            if not os.path.isdir(self.targets_folder):
                 print(self.input_file)
                 print("given image location is not existing.")
                 exit()
@@ -147,6 +168,8 @@ class Clickomat:
 
         self.logging              = True
         self.step_pause           = 0.06
+        self.dflt_drag_duration   = 0.3
+        self.dflt_move_duration   = 0.03
         self.switch_pause         = 0
         self.switched             = 0
         self.switch               = True
@@ -154,11 +177,21 @@ class Clickomat:
         # Flags
         self.breakout             = False
         self.stopped              = False
+        self.finished             = False
         self.panicked             = False
         self.error                = ""
         self.test                 = False # needed for pytest
 
     # endregion
+    # region panic
+    def _panic(self) -> None:
+        print("\nPanic Stopp!!!")
+        self.error    = "Panic Stopp!!!"
+        self.panicked = True
+        self.stopped  = True
+        self.breakout = True
+        self._stopAllTreads()
+    # endregion
     # region _pause(line)
     def _pause(self,line):
         # pause method is run before every command
@@ -168,20 +201,22 @@ class Clickomat:
         # when False is returned the actual pause in the clickloop is skipped
         def sp():
             if self.step_pause > 0:
-                return float(self.step_pause)
+                result = [float(self.step_pause),False]  # True means: this command line was NOT a pause!
+                return result
             else:
-                return False
+                return False  # No tuple / only False means: No pause at all. Not even default pause
         if line == ".": sp()
         pause = re.search(r"^[0-9]+(\.?[0-9]+)?$", line)
         if pause:
             pause = pause.group(0)
-            return float(pause)
+            result = [float(pause),True]  # True means: this command line WAS a pause!
+            return result
         else:
             return sp()
 
     # endregion
-    # region _getImage(line)
-    def _getImage(self,line):
+    # region _getImages(line)
+    def _getImages(self,line):
         # looks if 1 or more images can be found in the command line
         # and returns a list with png images
         # if no image(s) found 'Click' is returned and that causes the
@@ -202,7 +237,7 @@ class Clickomat:
             img = img.split("/")
 
             for i in img:
-                this = f"{self.images}/{i}.png"
+                this = f"{self.targets_folder}/{i}.png"
                 if exists(this):
                     result.append(this)
 
@@ -260,16 +295,16 @@ class Clickomat:
     # region _getTimeout(line)
     def _getTimeout(self,line,default=10):
         try:
-            timeout = re.search(r" [0-9]+ ", line).group(0)
-            return int(timeout)
+            timeout = re.search(r" [0-9\.]+ ", line).group(0)
+            return float(timeout)
         except:
             return default
     # endregion
-    # region _findImage(image)
-    def _findImage(self,image):
+    # region _findImages(image)
+    def _findImages(self,images):
         # takes LIST of images!
         x = False
-        for i in image:
+        for i in images:
             try:
                 x,_ = pyautogui.locateCenterOnScreen(i, confidence=self.confidence)
             except: pass
@@ -310,21 +345,24 @@ class Clickomat:
     # region _switch()
     def _switch(self):
     # no unittest for this
-        if os.name == 'nt':
-            if not self.test: pyautogui.keyDown('alt')
-            if not self.test: pyautogui.press('tab')
-            if not self.test: pyautogui.keyUp('alt')
-        else:
-            if not self.test: pyautogui.keyDown('command')
-            if not self.test: pyautogui.press('tab')
-            if not self.test: pyautogui.keyUp('command')
+        if not self.test:
+            if os.name == 'nt':
+                kb.press(Key.alt)
+                kb.press(Key.tab)
+                kb.release(Key.tab)
+                kb.release(Key.alt)
+            else:
+                kb.press(Key.cmd)
+                kb.press(Key.tab)
+                kb.release(Key.tab)
+                kb.release(Key.cmd)
+
         if self.switch_pause > 0:
             time.sleep(self.switch_pause)
     # endregion
     # region _write(line)
     def _write(self,line):
         try:
-            # text = re.search(r" \"[a-zA-Z0-9_:@\-\.\/\\ ]+\"", line).group(0)
             text = re.search(r" \".+\"", line).group(0)
             text = text[2:len(text)-1]
             if self.logging: print(" -> ", text, end = "")
@@ -358,31 +396,31 @@ class Clickomat:
     def _push(self,order):
         amount = int(order[1])
         if order[0] == "up":
-            if not self.test: pyautogui.moveRel(0,amount*-1)
+            if not self.test: pyautogui.moveRel(0,amount*-1,duration=self.dflt_move_duration)
             return 'up'
         if order[0] == "down":
-            if not self.test: pyautogui.moveRel(0,amount)
+            if not self.test: pyautogui.moveRel(0,amount,duration=self.dflt_move_duration)
             return 'down'
         if order[0] == "right":
-            if not self.test: pyautogui.moveRel(amount,0)
+            if not self.test: pyautogui.moveRel(amount,0,duration=self.dflt_move_duration)
             return 'right'
         if order[0] == "left":
-            if not self.test: pyautogui.moveRel(amount*-1,0)
+            if not self.test: pyautogui.moveRel(amount*-1,0,duration=self.dflt_move_duration)
             return 'left'
         return False
     # endregion
     # region _click(line,mode)
     def _click(self,line,mode):
-            image = self._getImage(line)
+            images = self._getImages(line)
 
-            if not image:
+            if not images:
                 self._imageNotFound()
                 return("imageNotFound")
 
-            if image != "Click":
-                image = self._findImage(image)
+            if images != "Click":
+                images = self._findImages(images)
 
-            if image == "Click":
+            if images == "Click":
                 if mode==1:
                     if not self.test: pyautogui.click()
                     if self.logging: print(" -> clicked!", end="")
@@ -398,7 +436,7 @@ class Clickomat:
                     if self.logging: print(" -> shift-clicked!", end="")
                     return("normalShiftClickExecuted")
 
-            if not self._clickImage(image,mode):
+            if not self._clickImage(images,mode):
                 if self.logging: print(" -> not clicked!", end="")
                 order = line.split(" ")
                 if " ! " in order:
@@ -422,8 +460,9 @@ class Clickomat:
     # endregion
     # region _pos(line)
     def _pos(self,line):
-        image = self._getImage(line)
-        try: image = image[0]
+        images = self._getImages(line)
+        image = False
+        try: image = images[0]
         except: pass
 
         if not image:
@@ -433,7 +472,7 @@ class Clickomat:
         box = self._locateImage(image)
         if box:
             if self.logging: print(" -> ", box, end = "")
-            if not self.test: pyautogui.moveTo((box[0]+(box[2]/2)),(box[1]+(box[3]/2)))
+            if not self.test: pyautogui.moveTo((box[0]+(box[2]/2)),(box[1]+(box[3]/2)),self.dflt_move_duration)
             return("positionedSuccessful")
         else:
             if self.logging: print(" -> Position not found!", end="")
@@ -453,7 +492,7 @@ class Clickomat:
             print("no value assigned!")
             return False
         try:
-            if not self.test: pyautogui.moveTo(x,y)
+            if not self.test: pyautogui.moveTo(x,y,self.dflt_move_duration)
             return True
         except:
             return False
@@ -469,12 +508,15 @@ class Clickomat:
     # endregion
     # region _drag(line)
     def _drag(self,line):
-        # TODO: strange behavior with set duration
         order = line.split(" ")
-        image = self._getImage(line)
-        dur   = self._getTimeout(line,1)
-        try: image = image[0]
+        image = False
+        images = self._getImages(line)
+        dur   = float(self._getTimeout(line,self.dflt_drag_duration))
+        print()
+        print(dur)
+        try: image = images[0]
         except: pass
+
         if not image:
             self._imageNotFound()
             return("imageNotFound")
@@ -483,14 +525,12 @@ class Clickomat:
         if box:
             if self.logging: print(" -> ", box, end = "")
             if "up" in order:
-                if not self.test: pyautogui.moveTo((box[0]),(box[1]+box[3]))
-                if not self.test: pyautogui.dragTo((box[0]+box[2]),(box[1]), button='left')
-                if not self.test and dur: pyautogui.dragTo((box[0]+box[2]),(box[1]), dur , button='left')
+                if not self.test: pyautogui.moveTo((box[0]),(box[1]+box[3]),self.dflt_move_duration)
+                if not self.test: pyautogui.dragTo((box[0]+box[2]),(box[1]),dur,button='left')
                 return ("dragUpSuccess")
             else:
-                if not self.test: pyautogui.moveTo(box[0],box[1])
-                if not self.test: pyautogui.dragTo((box[0]+box[2]),(box[1]+box[3]), button='left')
-                if not self.test and dur: pyautogui.dragTo((box[0]+box[2]),(box[1]+box[3]), dur , button='left')
+                if not self.test: pyautogui.moveTo(box[0],box[1],self.dflt_move_duration)
+                if not self.test: pyautogui.dragTo((box[0]+box[2]),(box[1]+box[3]),dur,button='left')
                 return ("dragSuccess")
         else:
             if self.logging: print(" -> nothing to drag!", end="")
@@ -506,12 +546,12 @@ class Clickomat:
         if self.logging: print(" -> timeout: " + str(timeout) + "s", end = "")
         start_time = datetime.now()
 
-        image = self._getImage(line)
-        if not image:
+        images = self._getImages(line)
+        if not images:
             self._imageNotFound()
             return("imageNotFound")
 
-        if self.logging: print(f" on {str(image)}", end = "")
+        if self.logging: print(f" on {str(images)}", end = "")
 
         while 1:
             self._stopLoop()
@@ -519,7 +559,7 @@ class Clickomat:
             if time_delta.total_seconds() >= timeout:
                 break
 
-            if self._findImage(image):
+            if self._findImages(images):
                 t=round(time_delta.total_seconds())
                 if self.logging: print(" -> found after " + str(t) + "s.",end = "")
                 found = True
@@ -565,12 +605,16 @@ class Clickomat:
     def _stopAllTreads(self):
         try: self.Lookup.stop()
         except: pass
-        try: self.Blacklist.stop()
-        except: pass
-        try: self.Whitelist.stop()
-        except: pass
-        try: self.Panic.stop()
-        except: pass
+
+        if self.breakout or self.stopped or self.finished:
+            try: self.Blacklist.stop()
+            except: pass
+            try: self.Whitelist.stop()
+            except: pass
+            try: self.Panic.stop()
+            except: pass
+            try: self.Stoploop.stop()
+            except: pass
     # endregion
     # region _stopLoop()
     def _stopLoop(self):
@@ -594,9 +638,9 @@ class Clickomat:
     def _setLookup(self,line):
         if self.logging: print("Set Lookup!", end=" -> ")
 
-        image = self._getImage(line)
+        images = self._getImages(line)
         sec   = self._getSection(line)
-        if image and sec: target = [image,sec]
+        if images and sec: target = [images,sec]
         if self.logging: print(target)
 
         try:
@@ -612,45 +656,53 @@ class Clickomat:
         return
     # endregion
     # region _setWatcher()
-    def _setWatcher(self,line,name="lookup"):
+    def _setWatcher(self,line,name="lookup") -> bool:
         if self.logging: print(f"Set watcher: {name}!", end=" -> ")
 
-        image = self._getImage(line)
+        images = self._getImages(line)
         sec   = self._getSection(line)
-        if image and sec: target = [image,sec]
+
+        if images and sec:
+            target = [images,sec]
+        else:
+            return False
+
         if self.logging: print(target)
 
         if name == "lookup":
             try:
                 self.Lookup.setTarget(target)
+                self.Lookup.start()
+                return True
             except:
                 self.Lookup = Watcher(self,target,"lookup")
                 self.Lookup.start()
-            return
+                return True
 
         if name == "blacklist":
             try:
                 self.Blacklist.setTarget(target)
+                return True
             except:
                 self.Blacklist = Watcher(self,target,"blacklist")
                 self.Blacklist.start()
-            return
+                return True
 
         if name == "whitelist":
             try:
                 self.Whitelist.setTarget(target)
+                return True
             except:
                 self.Whitelist = Watcher(self,target,"whitelist")
                 self.Whitelist.start()
-            return
-
+                return True
 
     # endregion
     # region _if(line)
     def _if(self,line):
-        image = self._getImage(line)
+        images = self._getImages(line)
         sec = self._getSection(line)
-        if self._findImage(image) and sec:
+        if self._findImages(images) and sec:
             if self.logging: print(f"\nImage (if) found -> Go Section {sec}!\n")
             self.section = sec
             if not self.test: self.ClickLoop(self.section)
@@ -787,7 +839,7 @@ class Clickomat:
 
         if self.logging: print(f"Case Path: {self.case_path}")
         if self.logging: print(f"Clicklist: {self.input_file}")
-        if self.logging: print(f"Image Directory: {self.images}")
+        if self.logging: print(f"Image Directory: {self.targets_folder}")
         if self.logging: print("---------------------------------------------------")
 
         self._getClicklist()
@@ -795,10 +847,12 @@ class Clickomat:
         if not self.test: pyautogui.PAUSE = 0
         if not self.test: pyautogui.FAILSAFE = False
 
-
         self.Panic = Failsave(self)
         self.Panic.start()
 
+        self.Stoploop = StopLoop(self)
+        self.Stoploop.start()
+
         self.ClickLoop(self.section)
 
         if not self.breakout:
@@ -806,7 +860,6 @@ class Clickomat:
             if self.logging: print ("Loop finished.\n\n")
     # endregion
     # region clickloop
-
     def ClickLoop(self,sec):
 
         try: self.Lookup.nullTarget()
@@ -816,24 +869,21 @@ class Clickomat:
 
         for line in self.sections[sec]:
 
-            # TODO: section control as Watcher?
             if sec != self.section: break
 
             lnr  = line[0]
             line = line[1]
 
             if self.logging: print()
-
-            p = self._pause(line)
-            if p: time.sleep(p)
-
-            # TODO: stoploop as Watcher?
-            self._stopLoop()
-
             if self.logging: print(lnr, end = " " )
             if self.logging: print(line, end = "" )
 
-            # Thre at...
+            p = self._pause(line)
+            if p: time.sleep(p[0])
+            if p[1]:
+                if self.logging: print(f" -> Pause: {p[0]} sec.", end = "")
+                continue
+
             if sec != self.section: break
 
             order   = line.split(" ")
@@ -841,19 +891,20 @@ class Clickomat:
 
             if command == "#": continue
 
-            # lookup is a Watcher now (>= v0.3.3)
+            # lookup is a Watcher now (>= v1.0.0)
             if command == "lookup" or command == "lu":
                 self._setWatcher(line,"lookup")
 
             # blacklist watcher
             if command == "blacklist" or command == "bl":
-                self._setWatcher(line,"blacklist")
+                result = self._setWatcher(line,"blacklist")
+                if not result: self._panic()
 
             # whitelist watcher
             if command == "whitelist" or command == "wl":
-                self._setWatcher(line,"whitelist")
+                result = self._setWatcher(line,"whitelist")
+                if not result: self._panic()
 
-            # Threat...
             if sec != self.section: break
 
             if command == "stop":
@@ -866,32 +917,24 @@ class Clickomat:
                     self.switched += 1
                 continue
 
-            # Threat...
             if sec != self.section: break
 
-            # Threat...
-            self._stopLoop()
-
             pushroute = self._pushRoute(command,order)
             if pushroute: eval(pushroute); continue
 
-            # Threat...
-            self._stopLoop()
-
             clickroute = self._clickRoute(command,line)
             if clickroute: eval(clickroute); continue
 
-            # Threat...
-            self._stopLoop()
-
             route = self._routes(command)
             if route: eval(route); continue
 
-            if command == "end": self._end()
+            if command == "end":
+                self._end()
+                continue
 
-            # Threat...
-            self._stopLoop()
+            if self.logging: print(" -> Command not recognized - check syntax.", end = "")
 
+        self.finished = True
         self._stopAllTreads()
 
         if self.autoswitch and self.switched == 1:
@@ -899,7 +942,6 @@ class Clickomat:
             self._switch()
 
     # endregion
-
 # region click arguments
 CONTEXT_SETTINGS = dict(help_option_names=['-h','-help','--help'],max_content_width=400)
 @click.command(context_settings=CONTEXT_SETTINGS)
@@ -912,19 +954,22 @@ CONTEXT_SETTINGS = dict(help_option_names=['-h','-help','--help'],max_content_wi
 @click.option('--silent', '-silent', '-s', flag_value=True, default=False, help='if set Clickomat gives no terminal feedback, default: False')
 @click.option('--step', '-step', '-st', default=0.06, type=click.FLOAT, help='set the pause length between each command, default: 0.06')
 @click.option('--noswitch','-noswitch', '-n', flag_value=True, default=False, help='if set Clickomat will not execute switch commands unless they are marked with `!`')
-def run(version,path,clicklist,images,confidence,autoswitch,silent,step,noswitch):
+@click.option('--position','-position', '-pos', flag_value=True, default=False, help='lets you copy X/Y coordinates for your clickscript')
+def run(version,path,clicklist,images,confidence,autoswitch,silent,step,noswitch,position):
 
     """Clickomat documentation is available under https://github.com/skilleven/clickomat/wiki"""
 
+    if position: clipPositionLoop()
+
     if version:
-        print("Clickomat 0.3.3 is installed.\nYou may want to check if your version is up to date: pip list --outdated")
+        print("Clickomat 1.0.2 is installed.\nYou may want to check if your version is up to date: pip list --outdated")
         exit()
 
     case_path = path
 
     # check if clicklist parameter is correct
     if not ".txt" in clicklist:
-        input_file = f"{clicklist}.txt"
+        clicklist = f"{clicklist}.txt"
     input_file = os.path.join(case_path, clicklist)
     if not os.path.exists(input_file):
         print(f"Specified clicklist is not existing: {input_file}")
@@ -943,10 +988,10 @@ def run(version,path,clicklist,images,confidence,autoswitch,silent,step,noswitch
         print("Confidence can have a maximum of 1.0!")
         exit()
 
+
     go(case_path,clicklist,images,confidence,autoswitch,silent,step,noswitch)
 # endregion
-
-#region go
+# region go
 def go(case_path,input_file,images,confidence,autoswitch,silent,step,noswitch):
     c = Clickomat(case_path,input_file,images)
     c.confidence = confidence
@@ -956,6 +1001,36 @@ def go(case_path,input_file,images,confidence,autoswitch,silent,step,noswitch):
     if noswitch: c.switch = False
     c.main()
 # endregion
+# region clipPositionLoop()
+def clipPositionLoop():
+    try:
+        oldpos = ''
+        while True:
+            pos = pyautogui.position()
+            spx = ""
+            if pos.x < 1000: spx = " "
+            if pos.x < 100: spx = "  "
+            if pos.x < 10: spx = "   "
+
+            printout = f"X {pos.x}{spx}      Y {pos.y}"
+            if (pos.x < 0) or (pos.y < 0): printout = "Only Main Screen is supported by clickomat."
+
+            clipboard = f"posX {pos.x}\nposY {pos.y}"
+            if printout != oldpos: print(printout)
+            oldpos = printout
+
+            # The event listener will be running in this block
+            with kbd.Events() as events:
+                # Block at most 0.1 second
+                event = events.get(0.1)
+                if event:
+                    pyperclip.copy(clipboard)
+                    exit()
+
+    except KeyboardInterrupt:
+        pyperclip.copy(clipboard)
+        exit()
+# endregion
 
 if __name__ == "__main__":
     run()
